# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.packages
~~~~~~~~~~~~~~~~

Third-party libs that we bundle with Warthog
"""



# -*- coding: utf-8 -*-

import pytest
import requests
import mock

import warthog.core
import warthog.exceptions


SOME_CRAZY_ERROR = {
    'response': {
        'status': 'fail',
        'err': {
            'code': 10001,
            'msg': 'You done did it now'
        }
    }
}

BAD_PW = {
    'response': {
        'status': 'fail',
        'err': {
            'code': 520486915,
            'msg': 'Bad password'
        }
    }
}

INVALID_SESSION = {
    'response': {
        'status': 'fail',
        'err': {
            'code': 1009,
            'msg': 'Invalid session'
        }
    }
}

NO_SUCH_SERVER = {
    'response': {
        'status': 'fail',
        'err': {
            'code': 67174402,
            'msg': 'That server does not exist'
        }
    }
}

OK_RESPONSE = {
    'response': {
        'status': 'OK'
    }
}

NODE_STATUS = {
    'server_stat': {
        'status': 1,
        'cur_conns': 42
    }
}

SCHEME_HOST = 'https://lb.example.com'


@pytest.fixture
def response():
    return mock.Mock(spec=requests.Response)


@pytest.fixture
def transport(response):
    mock_transport = mock.Mock(spec=requests.Session)
    mock_transport.get.return_value = response
    mock_transport.post.return_value = response
    return mock_transport


class TestSessionStartCommand(object):
    def test_send_bad_password(self, transport, response):
        response.text = ''
        response.json.return_value = dict(BAD_PW)

        with pytest.raises(warthog.exceptions.WarthogAuthFailureError):
            cmd = warthog.core.SessionStartCommand(transport, SCHEME_HOST, 'user', 'bad password')
            cmd.send()

        assert transport.post.called, 'Expected transport ".get()" to be called'

    def test_send_success(self, transport, response):
        response.text = ''
        response.json.return_value = {
            'session_id': '1234'
        }

        cmd = warthog.core.SessionStartCommand(transport, SCHEME_HOST, 'user', 'password')
        session = cmd.send()

        assert '1234' == session, 'Did not get expected session ID'


class TestSessionEndCommand(object):
    def test_send_invalid_session(self, transport, response):
        response.text = ''
        response.json.return_value = dict(INVALID_SESSION)

        with pytest.raises(warthog.exceptions.WarthogInvalidSessionError):
            cmd = warthog.core.SessionEndCommand(transport, SCHEME_HOST, 'bad session')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_unknown_error(self, transport, response):
        response.text = ''
        response.json.return_value = dict(SOME_CRAZY_ERROR)

        with pytest.raises(warthog.exceptions.WarthogAuthCloseError):
            cmd = warthog.core.SessionEndCommand(transport, SCHEME_HOST, '1234')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_success(self, transport, response):
        response.text = ''
        response.json.return_value = dict(OK_RESPONSE)

        cmd = warthog.core.SessionEndCommand(transport, SCHEME_HOST, '1234')
        closed = cmd.send()
        assert closed, 'Did not get expected True result from session close'
        assert transport.post.called, 'Expected transport ".post() to be called'


class TestNodeEnableCommand(object):

    def test_send_invalid_session(self, transport, response):
        response.text = ''
        response.json.return_value = dict(INVALID_SESSION)

        with pytest.raises(warthog.exceptions.WarthogInvalidSessionError):
            cmd = warthog.core.NodeEnableCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_no_such_server(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NO_SUCH_SERVER)

        with pytest.raises(warthog.exceptions.WarthogNoSuchNodeError):
            cmd = warthog.core.NodeEnableCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_unknown_error(self, transport, response):
        response.text = ''
        response.json.return_value = dict(SOME_CRAZY_ERROR)

        with pytest.raises(warthog.exceptions.WarthogNodeEnableError):
            cmd = warthog.core.NodeEnableCommand(
                transport, SCHEME_HOST, '1234', 'good.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_success(self, transport, response):
        response.text = ''
        response.json.return_value = dict(OK_RESPONSE)

        cmd = warthog.core.NodeEnableCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        got_enabled = cmd.send()
        assert got_enabled, 'Did not get get expected True result from node enable'
        assert transport.post.called, 'Expected transport ".post() to be called'


class TestNodeDisableCommand(object):
    def test_send_invalid_session(self, transport, response):
        response.text = ''
        response.json.return_value = dict(INVALID_SESSION)

        with pytest.raises(warthog.exceptions.WarthogInvalidSessionError):
            cmd = warthog.core.NodeDisableCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_no_such_server(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NO_SUCH_SERVER)

        with pytest.raises(warthog.exceptions.WarthogNoSuchNodeError):
            cmd = warthog.core.NodeDisableCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_unknown_error(self, transport, response):
        response.text = ''
        response.json.return_value = dict(SOME_CRAZY_ERROR)

        with pytest.raises(warthog.exceptions.WarthogNodeDisableError):
            cmd = warthog.core.NodeDisableCommand(
                transport, SCHEME_HOST, '1234', 'good.example.com')
            cmd.send()

        assert transport.post.called, 'Expected transport ".post() to be called'

    def test_send_success(self, transport, response):
        response.text = ''
        response.json.return_value = dict(OK_RESPONSE)

        cmd = warthog.core.NodeDisableCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        got_disabled = cmd.send()
        assert got_disabled, 'Did not get get expected True result from node disable'
        assert transport.post.called, 'Expected transport ".post() to be called'


class TestNodeStatusCommand(object):
    def test_send_invalid_session(self, transport, response):
        response.text = ''
        response.json.return_value = dict(INVALID_SESSION)

        with pytest.raises(warthog.exceptions.WarthogInvalidSessionError):
            cmd = warthog.core.NodeStatusCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_no_such_server(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NO_SUCH_SERVER)

        with pytest.raises(warthog.exceptions.WarthogNoSuchNodeError):
            cmd = warthog.core.NodeStatusCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_unknown_error(self, transport, response):
        response.text = ''
        response.json.return_value = dict(SOME_CRAZY_ERROR)

        with pytest.raises(warthog.exceptions.WarthogNodeStatusError):
            cmd = warthog.core.NodeStatusCommand(
                transport, SCHEME_HOST, '1234', 'good.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_server_enabled(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NODE_STATUS)

        cmd = warthog.core.NodeStatusCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        status = cmd.send()
        assert warthog.core.STATUS_ENABLED == status, 'Did not get expected enabled status'
        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_server_disabled(self, transport, response):
        payload = dict(NODE_STATUS)
        payload['server_stat']['status'] = 0

        response.text = ''
        response.json.return_value = payload

        cmd = warthog.core.NodeStatusCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        status = cmd.send()
        assert warthog.core.STATUS_DISABLED == status, 'Did not get expected disabled status'
        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_server_down(self, transport, response):
        payload = dict(NODE_STATUS)
        payload['server_stat']['status'] = 2

        response.text = ''
        response.json.return_value = payload

        cmd = warthog.core.NodeStatusCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        status = cmd.send()
        assert warthog.core.STATUS_DOWN == status, 'Did not get expected down status'
        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_server_no_known_status(self, transport, response):
        payload = dict(NODE_STATUS)
        payload['server_stat']['status'] = 999

        response.text = ''
        response.json.return_value = payload

        with pytest.raises(warthog.exceptions.WarthogNodeStatusError):
            cmd = warthog.core.NodeStatusCommand(
                transport, SCHEME_HOST, '1234', 'good.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'


class TestNodeActiveConnectionsCommand(object):
    def test_send_invalid_session(self, transport, response):
        response.text = ''
        response.json.return_value = dict(INVALID_SESSION)

        with pytest.raises(warthog.exceptions.WarthogInvalidSessionError):
            cmd = warthog.core.NodeActiveConnectionsCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_no_such_server(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NO_SUCH_SERVER)

        with pytest.raises(warthog.exceptions.WarthogNoSuchNodeError):
            cmd = warthog.core.NodeActiveConnectionsCommand(
                transport, SCHEME_HOST, '1234', 'bad.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_unknown_error(self, transport, response):
        response.text = ''
        response.json.return_value = dict(SOME_CRAZY_ERROR)

        with pytest.raises(warthog.exceptions.WarthogNodeStatusError):
            cmd = warthog.core.NodeActiveConnectionsCommand(
                transport, SCHEME_HOST, '1234', 'good.example.com')
            cmd.send()

        assert transport.get.called, 'Expected transport ".get() to be called'

    def test_send_success(self, transport, response):
        response.text = ''
        response.json.return_value = dict(NODE_STATUS)

        cmd = warthog.core.NodeActiveConnectionsCommand(
            transport, SCHEME_HOST, '1234', 'good.example.com')
        connections = cmd.send()
        assert 42 == connections, 'Did not get expected active connections'
        assert transport.get.called, 'Expected transport ".get() to be called'




# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.client
~~~~~~~~~~~~~~

Simple interface for a load balancer with retry logic and intelligent draining of nodes.
"""

import time
import contextlib

import warthog.core
import warthog.exceptions
import warthog.transport


class CommandFactory(object):
    """Factory for getting new :mod:`warthog.core` command instances that each
    perform some type of request against the load balancer API.

    It is typically not required for user code to instantiate this class directly
    unless you have special requirements and need to inject a custom ``transport_factory``
    method.

    This class is thread safe.
    """

    def __init__(self, transport_factory):
        """Set the a factory that will create new HTTP Sessions instances to be
        used for executing commands.

        :param callable transport_factory: Callable for creating new Session instances
            for executing commands.
        """
        self._transport_factory = transport_factory

    def get_session_start(self, scheme_host, username, password):
        """Get a new command instance to start a session.

        :param basestring scheme_host: Scheme, host, and port combination of the load balancer.
        :param basestring username: Name of the user to authenticate with.
        :param basestring password: Password for the user to authenticate with.
        :return: A new command to start a session.
        :rtype: warthog.core.SessionStartCommand
        """
        return warthog.core.SessionStartCommand(
            self._transport_factory(), scheme_host, username, password)

    def get_session_end(self, scheme_host, session_id):
        """Get a new command instance to close an existing session.

        :param basestring scheme_host: Scheme, host, and port combination of
            the load balancer.
        :param basestring session_id: Previously authenticated session ID.
        :return: A new command to close a session.
        :rtype: warthog.core.SessionEndCommand
        """
        return warthog.core.SessionEndCommand(
            self._transport_factory(), scheme_host, session_id)

    def get_server_status(self, scheme_host, session_id, server):
        """Get a new command to get the status (enabled / disabled) of a server.

        :param basestring scheme_host: Scheme, host, and port combination of
            the load balancer.
        :param basestring session_id: Previously authenticated session ID.
        :param basestring server: Host name of the server to get the status of.
        :return: A new command to get the status of a server.
        :rtype: warthog.core.NodeStatusCommand
        """
        return warthog.core.NodeStatusCommand(
            self._transport_factory(), scheme_host, session_id, server)

    def get_enable_server(self, scheme_host, session_id, server):
        """Get a new command to enable a server at the node level.

        :param basestring scheme_host: Scheme, host, and port combination of
            the load balancer.
        :param basestring session_id: Previously authenticated session ID.
        :param basestring server: Host name of the server to enable.
        :return: A new command to enable a server.
        :rtype: warthog.core.NodeEnableCommand
        """
        return warthog.core.NodeEnableCommand(
            self._transport_factory(), scheme_host, session_id, server)

    def get_disable_server(self, scheme_host, session_id, server):
        """Get a new command to disable a server at the node level.

        :param basestring scheme_host: Scheme, host, and port combination of the
            load balancer.
        :param basestring session_id: Previously authenticated session ID.
        :param basestring server: Host name of the server to disable.
        :return: A new command to disable a server.
        :rtype: warthog.core.NodeDisableCommand
        """
        return warthog.core.NodeDisableCommand(
            self._transport_factory(), scheme_host, session_id, server)

    def get_active_connections(self, scheme_host, session_id, server):
        """Get a new command to get the number of active connections to a server.

        :param basestring scheme_host: Scheme, host, and port combination of
            the load balancer.
        :param basestring session_id: Previously authenticated session ID.
        :param basestring server: Host name of the server to get the number of
            active connections to.
        :return: A new command to get active connections to a server.
        :rtype: warthog.core.NodeActiveConnectionsCommand
        """
        return warthog.core.NodeActiveConnectionsCommand(
            self._transport_factory(), scheme_host, session_id, server)


def _get_default_cmd_factory(verify, ssl_version):
    """Get a :class:`CommandFactory` instance configured to use the provided TLS
    version and cert verification policy

    :param bool verify: ``True`` to perform certificate validation when using HTTPS,
        ``False`` otherwise, ``None`` to use the default.
    :param int ssl_version: :mod:`ssl` module constant for specifying which SSL or
        TLS version to use for connecting to the load balancer over HTTPS, ``None``
        to use the default.
    :return: Default command factory for building new commands to interact
        with the A10 load balancer.
    :rtype: WarthogCommandFactory
    """
    return CommandFactory(warthog.transport.get_transport_factory(
        verify=verify, ssl_version=ssl_version))


@contextlib.contextmanager
def session_context(scheme_host, username, password, commands):
    """Context manager that makes a request to start an authenticated session, yields the
    session ID, and then closes the session afterwards.

    :param basestring scheme_host: Scheme, host, and port combination of the load balancer.
    :param basestring username: Name of the user to authenticate with.
    :param basestring password: Password for the user to authenticate with.
    :param CommandFactory commands: Factory instance for creating new commands
        for starting and ending sessions with the load balancer.
    :return: The session ID of the newly established session.
    """
    session = None
    try:
        start_cmd = commands.get_session_start(scheme_host, username, password)
        session = start_cmd.send()

        yield session
    finally:
        if session is not None:
            end_cmd = commands.get_session_end(scheme_host, session)
            end_cmd.send()


class WarthogClient(object):
    """Client for interacting with an A10 load balancer to get the status
    of nodes managed by it, enable them, and disable them.

    This class is thread safe.

    .. versionchanged:: 0.8.0
        Removed .disabled_context() method.
    """
    _logger = warthog.core.get_log()
    _default_wait_interval = 2.0

    # pylint: disable=too-many-arguments
    def __init__(self, scheme_host, username, password,
                 verify=None,
                 ssl_version=None,
                 wait_interval=_default_wait_interval,
                 commands=None):
        """Set the load balancer scheme/host/port combination, username and password
        to use for connecting and authenticating with the load balancer.

        Optionally, whether or not to verify certificates when using HTTPS may be
        toggled. This can enable you to use a self signed certificate for the load
        balancer while still using HTTPS.

        Optionally, the version of SSL or TLS to use may be specified as a :mod:`ssl`
        module protocol constant.

        Optionally, the amount of time to wait between retries of various operations
        and the factory used for creating commands may be set. If the interval between
        retries is not supplied, the default is two seconds.

        If the command factory is not supplied, a default instance will be used. The
        command factory is responsible for creating new :class:`requests.Session` instances
        to be used by each command. It is typically only necessary to override this for
        unit testing purposes.

        .. versionchanged:: 0.9.0
            Added the optional ``verify`` parameter to make use of self-signed certs
            easier.

        .. versionchanged:: 0.10.0
            Added the optional ``ssl_version`` parameter to make use of alternate SSL
            or TLS versions easier.

        :param basestring scheme_host: Scheme, host, and port combination of the load balancer.
        :param basestring username: Name of the user to authenticate with.
        :param basestring password: Password for the user to authenticate with.
        :param bool|None verify: ``True`` to verify certificates when using HTTPS, ``False``
            to skip verification, ``None`` to use the library default. The default is to
            verify certificates.
        :param int|None ssl_version: :mod:`ssl` module constant for specifying which version of
            SSL or TLS to use when connecting to the load balancer over HTTPS, ``None`` to use
            the library default. The default is to use TLSv1.
        :param float wait_interval: How long (in seconds) to wait between each retry of
            various operations (waiting for nodes to transition, waiting for connections
            to close, etc.).
        :param CommandFactory commands: Factory instance for creating new commands for
            starting and ending sessions with the load balancer.
        """
        self._scheme_host = scheme_host
        self._username = username
        self._password = password
        self._interval = wait_interval
        self._commands = commands if commands is not None else \
            _get_default_cmd_factory(verify, ssl_version)

    def _session_context(self):
        """Get a new context manager that starts and ends a session with the load balancer."""
        self._logger.debug('Creating new session context for %s', self._scheme_host)
        return session_context(self._scheme_host, self._username, self._password, self._commands)

    def get_status(self, server):
        """Get the current status of the given server, at the node level.

        The status will be one of the constants :data:`warthog.core.STATUS_ENABLED`
        :data:`warthog.core.STATUS_DISABLED`, or :data:`warthog.core.STATUS_DOWN`.

        :param basestring server: Hostname of the server to get the status of.
        :return: The current status of the server, enabled, disabled, or down.
        :rtype: basestring
        :raises warthog.exceptions.WarthogAuthFailureError: If authentication with
            the load balancer failed when trying to establish a new session for this
            operation.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the load balancer does
            not recognize the given hostname.
        :raises warthog.exceptions.WarthogNodeStatusError: If there are any other
            problems getting the status of the given server.
        """
        with self._session_context() as session:
            cmd = self._commands.get_server_status(
                self._scheme_host, session, server)
            return cmd.send()

    def get_connections(self, server):
        """Get the current number of active connections to a server, at the node level.

        The number of connections will be 0 or a positive integer.

        :param basestring server: Hostname of the server to get the number of active
            connections for.
        :return: The number of active connections total for the node, across all groups
            the server is in.
        :rtype: int
        :raises warthog.exceptions.WarthogAuthFailureError: If authentication with
            the load balancer failed when trying to establish a new session for this
            operation.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the load balancer does
            not recognize the given hostname.
        :raises warthog.exceptions.WarthogNodeStatusError: If there are any other
            problems getting the active connections for the given server.

        .. versionadded:: 0.4.0
        """
        with self._session_context() as session:
            cmd = self._commands.get_active_connections(
                self._scheme_host, session, server)
            return cmd.send()

    def disable_server(self, server, max_retries=5):
        """Disable a server at the node level, optionally retrying when there are transient
        errors and waiting for the number of active connections to the server to reach zero.

        If ``max_retries`` is zero, no attempt will be made to retry on transient errors
        or to wait until there are no active connections to the server, the method will
        try a single time to disable the server and then return immediately.

        :param basestring server: Hostname of the server to disable
        :param int max_retries: Max number of times to sleep and retry when encountering
            some sort of transient error when disabling the server and while waiting for
            the number of active connections to a server to reach zero.
        :return: True if the server was disabled, false otherwise.
        :rtype: bool
        :raises warthog.exceptions.WarthogAuthFailureError: If authentication with
            the load balancer failed when trying to establish a new session for this
            operation.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the load balancer does
            not recognize the given hostname.
        :raises warthog.exceptions.WarthogNodeDisableError: If there are any other
            problems disabling the given server.
        """
        with self._session_context() as session:
            disable = self._commands.get_disable_server(
                self._scheme_host, session, server)
            self._try_repeatedly(disable.send, max_retries)

            active = self._commands.get_active_connections(
                self._scheme_host, session, server)
            self._wait_for_connections(active.send, max_retries)

            status = self._commands.get_server_status(
                self._scheme_host, session, server)
            return warthog.core.STATUS_DISABLED == status.send()

    # NOTE: there's a fair amount of duplicate code between this method and _wait_for_status
    # and we could consolidate them to one method that just accepts a function and waits for
    # it to return true and then break. But, this way we have more useful debug information
    # logged at the expense of duplicate code.
    # pylint: disable=missing-docstring
    def _wait_for_connections(self, conn_method, max_retries):
        retries = 0

        while retries < max_retries:
            conns = conn_method()
            if conns == 0:
                break

            self._logger.debug(
                "Connections still active: %s, sleeping for %s seconds...", conns, self._interval)
            time.sleep(self._interval)
            retries += 1

    def enable_server(self, server, max_retries=5):
        """Enable a server at the node level, optionally retrying when there are transient
        errors and waiting for the server to enter the expected, enabled state.

        If ``max_retries`` is zero, no attempt will be made to retry on transient errors
        or to wait until the server enters the expected, enabled state, the method will
        try a single time to enable the server then return immediately.

        :param basestring server: Hostname of the server to enable
        :param int max_retries: Max number of times to sleep and retry when encountering
            some transient error while trying to enable the server
        :return: True if the server was enabled, false otherwise
        :rtype: bool
        :raises warthog.exceptions.WarthogAuthFailureError: If authentication with
            the load balancer failed when trying to establish a new session for this
            operation.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the load balancer does
            not recognize the given hostname.
        :raises warthog.exceptions.WarthogNodeEnableError: If there are any other
            problems enabling the given server.
        """
        with self._session_context() as session:
            enable = self._commands.get_enable_server(
                self._scheme_host, session, server)
            self._try_repeatedly(enable.send, max_retries)

            status = self._commands.get_server_status(
                self._scheme_host, session, server)
            self._wait_for_enable(status.send, max_retries)

            return warthog.core.STATUS_ENABLED == status.send()

    # pylint: disable=missing-docstring
    def _wait_for_enable(self, status_method, max_retries):
        retries = 0

        while retries < max_retries:
            status = status_method()
            if status == warthog.core.STATUS_ENABLED:
                break

            self._logger.debug(
                "Server is not yet enabled (%s), sleeping for %s seconds...",
                status, self._interval)
            time.sleep(self._interval)
            retries += 1

    def _try_repeatedly(self, method, max_retries):
        """Execute a method, retrying if it fails due to a transient error
        up to a given number of times, with the instance-wide interval in
        between each try.
        """
        retries = 0

        while True:
            try:
                return method()
            except warthog.exceptions.WarthogApiError as e:
                if e.api_code not in warthog.core.TRANSIENT_ERRORS or retries >= max_retries:
                    raise
                self._logger.debug(
                    "Encountered transient error %s - %s, retrying... ", e.api_code, e.api_msg)
                time.sleep(self._interval)
                retries += 1



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.transport
~~~~~~~~~~~~~~~~~

Methods to configure how to interact with the load balancer API over HTTP or HTTPS.
"""

import ssl

import warnings
import requests
from requests.adapters import (
    HTTPAdapter,
    DEFAULT_POOLBLOCK,
    DEFAULT_POOLSIZE,
    DEFAULT_RETRIES)
from requests.packages.urllib3.poolmanager import PoolManager
from requests.packages.urllib3.exceptions import InsecureRequestWarning


# Default to using the SSL/TLS version that the A10 requires instead of
# the default that the requests/urllib3 library picks. Or, maybe the A10
# just doesn't allow the client to negotiate. Either way, we use TLSv1.
# pylint: disable=no-member
DEFAULT_SSL_VERSION = ssl.PROTOCOL_TLSv1

# Default to verifying SSL/TLS certs because "safe by default" is a good idea.
DEFAULT_CERT_VERIFY = True


def get_transport_factory(verify=None, ssl_version=None):
    """Get a new callable that returns :class:`requests.Session` instances that
    have been configured according to the given parameters.

    :class:`requests.Session` instances are then used for interacting with the API
    of the load balancer over HTTP or HTTPS.

    It is typically not required for user code to call this function directly unless
    you have special requirements such as needing to bypass HTTPS certificate validation
    because you use a self signed certificate.

    .. versionchanged:: 0.10.0
        Using the requests/urllib3 default is no longer an option. Passing a ``None`` value
        for ``ssl_version`` will result in using the Warthog default (TLS v1).

    :param bool|None verify: Should SSL certificates by verified when connecting
        over HTTPS? Default is ``True``. If you have chosen not to verify certificates
        warnings about this emitted by the requests library will be suppressed.
    :param int|None ssl_version: Explicit version of SSL to use for HTTPS connections
        to an A10 load balancer. The version is a constant as specified by the
        :mod:`ssl` module. The default is TLSv1.
    :return: A callable to return new configured session instances for making HTTP(S)
        requests
    :rtype: callable
    """
    # Using `None` here to represent "not specified" so we don't have to litter the
    # whole lib with references to the default values we've specified here. Callers
    # just pass `None` and we'll pick the default here.
    verify = verify if verify is not None else DEFAULT_CERT_VERIFY
    ssl_version = ssl_version if ssl_version is not None else DEFAULT_SSL_VERSION

    # pylint: disable=missing-docstring
    def factory():
        transport = requests.Session()
        transport.mount('https://', VersionedSSLAdapter(ssl_version))

        if not verify:
            transport.verify = False

        return transport

    # Make sure that we suppress warnings about invalid certs since the user
    # has explicitly asked us to not verify it, they know that we're doing
    # something dangerous and don't care.
    if not verify:
        warnings.filterwarnings("ignore", category=InsecureRequestWarning)

    return factory


class VersionedSSLAdapter(HTTPAdapter):
    """"Transport adapter that requires the use of a specific version of SSL."""

    # pylint: disable=too-many-arguments
    def __init__(self, ssl_version, pool_connections=DEFAULT_POOLSIZE,
                 pool_maxsize=DEFAULT_POOLSIZE, max_retries=DEFAULT_RETRIES,
                 pool_block=DEFAULT_POOLBLOCK):
        self.ssl_version = ssl_version

        super(VersionedSSLAdapter, self).__init__(
            pool_connections=pool_connections,
            pool_maxsize=pool_maxsize,
            max_retries=max_retries,
            pool_block=pool_block)

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        # pylint: disable=attribute-defined-outside-init
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize, block=block,
            ssl_version=self.ssl_version, **pool_kwargs)



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog
~~~~~~~

Simple client for A10 load balancers.
"""

__version__ = '1.1.0'



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.exceptions
~~~~~~~~~~~~~~~~~~

Exceptions raised by the Warthog client or library.
"""


class WarthogError(Exception):
    """Base for all errors raised by the Warthog library."""

    def __init__(self, msg):
        super(WarthogError, self).__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class WarthogConfigError(WarthogError):
    """Base for errors raised while parsing or loading configuration."""


class WarthogNoConfigFileError(WarthogConfigError):
    """No configuration file could be found."""

    def __init__(self, msg, locations_checked=None):
        super(WarthogNoConfigFileError, self).__init__(msg)
        self.locations_checked = list(locations_checked) if locations_checked is not None else []

    def __str__(self):
        out = [self.msg]
        if self.locations_checked is not None:
            out.append('Locations checked: ' + ', '.join(self.locations_checked))
        return '. '.join(out)


class WarthogMalformedConfigFileError(WarthogConfigError):
    """The configuration file is missing required sections or fields."""

    def __init__(self, msg, missing_section=None, missing_option=None):
        super(WarthogMalformedConfigFileError, self).__init__(msg)
        self.missing_section = missing_section
        self.missing_option = missing_option

    def __str__(self):
        out = [self.msg]
        if self.missing_section is not None:
            out.append('Missing-section: {0}'.format(self.missing_section))
        if self.missing_option is not None:
            out.append('Missing-option: {0}'.format(self.missing_option))
        return '. '.join(out)


class WarthogApiError(WarthogError):
    """Base for errors raised in the course of interacting with the load balancer."""

    def __init__(self, msg, api_msg=None, api_code=None):
        super(WarthogApiError, self).__init__(msg)
        self.api_msg = api_msg
        self.api_code = api_code

    def __str__(self):
        out = [self.msg]
        if self.api_msg is not None:
            # Some error messages from the A10 end with a period, others don't
            out.append('API-message: {0}'.format(self.api_msg.rstrip('.')))
        if self.api_code is not None:
            out.append('API-code: {0}'.format(self.api_code))
        return '. '.join(out)


class WarthogAuthFailureError(WarthogApiError):
    """The credentials for authentication are invalid."""


class WarthogInvalidSessionError(WarthogApiError):
    """The session ID used while performing some action is unrecognized."""


class WarthogAuthCloseError(WarthogApiError):
    """There was some error while trying to end a session."""


class WarthogNodeError(WarthogApiError):
    """Base for errors specific to operating on some individual node."""

    def __init__(self, msg, api_msg=None, api_code=None, server=None):
        super(WarthogNodeError, self).__init__(msg, api_msg=api_msg, api_code=api_code)
        self.server = server


class WarthogNoSuchNodeError(WarthogNodeError):
    """The host being operated on is unrecognized."""


class WarthogNodeStatusError(WarthogNodeError):
    """There was some error while getting the status of a node."""


class WarthogNodeEnableError(WarthogNodeError):
    """There was some error while trying to enable a node."""


class WarthogNodeDisableError(WarthogNodeError):
    """There was some error while trying to disable a node."""



"""Utilities for writing code that runs on Python 2 and 3"""

# Copyright (c) 2010-2014 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: skip-file

from __future__ import absolute_import

import functools
import operator
import sys
import types

__author__ = "Benjamin Peterson <benjamin@python.org>"
__version__ = "1.8.0"


# Useful for very coarse version differentiation.
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes

    MAXSIZE = sys.maxsize
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    binary_type = str

    if sys.platform.startswith("java"):
        # Jython always uses 32 bits.
        MAXSIZE = int((1 << 31) - 1)
    else:
        # It's possible to have sizeof(long) != sizeof(Py_ssize_t).
        class X(object):
            def __len__(self):
                return 1 << 31
        try:
            len(X())
        except OverflowError:
            # 32-bit
            MAXSIZE = int((1 << 31) - 1)
        else:
            # 64-bit
            MAXSIZE = int((1 << 63) - 1)
        del X


def _add_doc(func, doc):
    """Add documentation to a function."""
    func.__doc__ = doc


def _import_module(name):
    """Import module, returning the module after the last dot."""
    __import__(name)
    return sys.modules[name]


class _LazyDescr(object):

    def __init__(self, name):
        self.name = name

    def __get__(self, obj, tp):
        result = self._resolve()
        setattr(obj, self.name, result) # Invokes __set__.
        # This is a bit ugly, but it avoids running this again.
        delattr(obj.__class__, self.name)
        return result


class MovedModule(_LazyDescr):

    def __init__(self, name, old, new=None):
        super(MovedModule, self).__init__(name)
        if PY3:
            if new is None:
                new = name
            self.mod = new
        else:
            self.mod = old

    def _resolve(self):
        return _import_module(self.mod)

    def __getattr__(self, attr):
        _module = self._resolve()
        value = getattr(_module, attr)
        setattr(self, attr, value)
        return value


class _LazyModule(types.ModuleType):

    def __init__(self, name):
        super(_LazyModule, self).__init__(name)
        self.__doc__ = self.__class__.__doc__

    def __dir__(self):
        attrs = ["__doc__", "__name__"]
        attrs += [attr.name for attr in self._moved_attributes]
        return attrs

    # Subclasses should override this
    _moved_attributes = []


class MovedAttribute(_LazyDescr):

    def __init__(self, name, old_mod, new_mod, old_attr=None, new_attr=None):
        super(MovedAttribute, self).__init__(name)
        if PY3:
            if new_mod is None:
                new_mod = name
            self.mod = new_mod
            if new_attr is None:
                if old_attr is None:
                    new_attr = name
                else:
                    new_attr = old_attr
            self.attr = new_attr
        else:
            self.mod = old_mod
            if old_attr is None:
                old_attr = name
            self.attr = old_attr

    def _resolve(self):
        module = _import_module(self.mod)
        return getattr(module, self.attr)


class _SixMetaPathImporter(object):
    """
    A meta path importer to import six.moves and its submodules.

    This class implements a PEP302 finder and loader. It should be compatible
    with Python 2.5 and all existing versions of Python3
    """
    def __init__(self, six_module_name):
        self.name = six_module_name
        self.known_modules = {}

    def _add_module(self, mod, *fullnames):
        for fullname in fullnames:
            self.known_modules[self.name + "." + fullname] = mod

    def _get_module(self, fullname):
        return self.known_modules[self.name + "." + fullname]

    def find_module(self, fullname, path=None):
        if fullname in self.known_modules:
            return self
        return None

    def __get_module(self, fullname):
        try:
            return self.known_modules[fullname]
        except KeyError:
            raise ImportError("This loader does not know module " + fullname)

    def load_module(self, fullname):
        try:
            # in case of a reload
            return sys.modules[fullname]
        except KeyError:
            pass
        mod = self.__get_module(fullname)
        if isinstance(mod, MovedModule):
            mod = mod._resolve()
        else:
            mod.__loader__ = self
        sys.modules[fullname] = mod
        return mod

    def is_package(self, fullname):
        """
        Return true, if the named module is a package.

        We need this method to get correct spec objects with
        Python 3.4 (see PEP451)
        """
        return hasattr(self.__get_module(fullname), "__path__")

    def get_code(self, fullname):
        """Return None

        Required, if is_package is implemented"""
        self.__get_module(fullname)  # eventually raises ImportError
        return None
    get_source = get_code  # same as get_code

_importer = _SixMetaPathImporter(__name__)


class _MovedItems(_LazyModule):
    """Lazy loading of moved objects"""
    __path__ = []  # mark as package


_moved_attributes = [
    MovedAttribute("cStringIO", "cStringIO", "io", "StringIO"),
    MovedAttribute("filter", "itertools", "builtins", "ifilter", "filter"),
    MovedAttribute("filterfalse", "itertools", "itertools", "ifilterfalse", "filterfalse"),
    MovedAttribute("input", "__builtin__", "builtins", "raw_input", "input"),
    MovedAttribute("intern", "__builtin__", "sys"),
    MovedAttribute("map", "itertools", "builtins", "imap", "map"),
    MovedAttribute("range", "__builtin__", "builtins", "xrange", "range"),
    MovedAttribute("reload_module", "__builtin__", "imp", "reload"),
    MovedAttribute("reduce", "__builtin__", "functools"),
    MovedAttribute("shlex_quote", "pipes", "shlex", "quote"),
    MovedAttribute("StringIO", "StringIO", "io"),
    MovedAttribute("UserDict", "UserDict", "collections"),
    MovedAttribute("UserList", "UserList", "collections"),
    MovedAttribute("UserString", "UserString", "collections"),
    MovedAttribute("xrange", "__builtin__", "builtins", "xrange", "range"),
    MovedAttribute("zip", "itertools", "builtins", "izip", "zip"),
    MovedAttribute("zip_longest", "itertools", "itertools", "izip_longest", "zip_longest"),

    MovedModule("builtins", "__builtin__"),
    MovedModule("configparser", "ConfigParser"),
    MovedModule("copyreg", "copy_reg"),
    MovedModule("dbm_gnu", "gdbm", "dbm.gnu"),
    MovedModule("_dummy_thread", "dummy_thread", "_dummy_thread"),
    MovedModule("http_cookiejar", "cookielib", "http.cookiejar"),
    MovedModule("http_cookies", "Cookie", "http.cookies"),
    MovedModule("html_entities", "htmlentitydefs", "html.entities"),
    MovedModule("html_parser", "HTMLParser", "html.parser"),
    MovedModule("http_client", "httplib", "http.client"),
    MovedModule("email_mime_multipart", "email.MIMEMultipart", "email.mime.multipart"),
    MovedModule("email_mime_nonmultipart", "email.MIMENonMultipart", "email.mime.nonmultipart"),
    MovedModule("email_mime_text", "email.MIMEText", "email.mime.text"),
    MovedModule("email_mime_base", "email.MIMEBase", "email.mime.base"),
    MovedModule("BaseHTTPServer", "BaseHTTPServer", "http.server"),
    MovedModule("CGIHTTPServer", "CGIHTTPServer", "http.server"),
    MovedModule("SimpleHTTPServer", "SimpleHTTPServer", "http.server"),
    MovedModule("cPickle", "cPickle", "pickle"),
    MovedModule("queue", "Queue"),
    MovedModule("reprlib", "repr"),
    MovedModule("socketserver", "SocketServer"),
    MovedModule("_thread", "thread", "_thread"),
    MovedModule("tkinter", "Tkinter"),
    MovedModule("tkinter_dialog", "Dialog", "tkinter.dialog"),
    MovedModule("tkinter_filedialog", "FileDialog", "tkinter.filedialog"),
    MovedModule("tkinter_scrolledtext", "ScrolledText", "tkinter.scrolledtext"),
    MovedModule("tkinter_simpledialog", "SimpleDialog", "tkinter.simpledialog"),
    MovedModule("tkinter_tix", "Tix", "tkinter.tix"),
    MovedModule("tkinter_ttk", "ttk", "tkinter.ttk"),
    MovedModule("tkinter_constants", "Tkconstants", "tkinter.constants"),
    MovedModule("tkinter_dnd", "Tkdnd", "tkinter.dnd"),
    MovedModule("tkinter_colorchooser", "tkColorChooser",
                "tkinter.colorchooser"),
    MovedModule("tkinter_commondialog", "tkCommonDialog",
                "tkinter.commondialog"),
    MovedModule("tkinter_tkfiledialog", "tkFileDialog", "tkinter.filedialog"),
    MovedModule("tkinter_font", "tkFont", "tkinter.font"),
    MovedModule("tkinter_messagebox", "tkMessageBox", "tkinter.messagebox"),
    MovedModule("tkinter_tksimpledialog", "tkSimpleDialog",
                "tkinter.simpledialog"),
    MovedModule("urllib_parse", __name__ + ".moves.urllib_parse", "urllib.parse"),
    MovedModule("urllib_error", __name__ + ".moves.urllib_error", "urllib.error"),
    MovedModule("urllib", __name__ + ".moves.urllib", __name__ + ".moves.urllib"),
    MovedModule("urllib_robotparser", "robotparser", "urllib.robotparser"),
    MovedModule("xmlrpc_client", "xmlrpclib", "xmlrpc.client"),
    MovedModule("xmlrpc_server", "SimpleXMLRPCServer", "xmlrpc.server"),
    MovedModule("winreg", "_winreg"),
    ]
for attr in _moved_attributes:
    setattr(_MovedItems, attr.name, attr)
    if isinstance(attr, MovedModule):
        _importer._add_module(attr, "moves." + attr.name)
del attr

_MovedItems._moved_attributes = _moved_attributes

moves = _MovedItems(__name__ + ".moves")
_importer._add_module(moves, "moves")


class Module_six_moves_urllib_parse(_LazyModule):
    """Lazy loading of moved objects in six.moves.urllib_parse"""


_urllib_parse_moved_attributes = [
    MovedAttribute("ParseResult", "urlparse", "urllib.parse"),
    MovedAttribute("SplitResult", "urlparse", "urllib.parse"),
    MovedAttribute("parse_qs", "urlparse", "urllib.parse"),
    MovedAttribute("parse_qsl", "urlparse", "urllib.parse"),
    MovedAttribute("urldefrag", "urlparse", "urllib.parse"),
    MovedAttribute("urljoin", "urlparse", "urllib.parse"),
    MovedAttribute("urlparse", "urlparse", "urllib.parse"),
    MovedAttribute("urlsplit", "urlparse", "urllib.parse"),
    MovedAttribute("urlunparse", "urlparse", "urllib.parse"),
    MovedAttribute("urlunsplit", "urlparse", "urllib.parse"),
    MovedAttribute("quote", "urllib", "urllib.parse"),
    MovedAttribute("quote_plus", "urllib", "urllib.parse"),
    MovedAttribute("unquote", "urllib", "urllib.parse"),
    MovedAttribute("unquote_plus", "urllib", "urllib.parse"),
    MovedAttribute("urlencode", "urllib", "urllib.parse"),
    MovedAttribute("splitquery", "urllib", "urllib.parse"),
    MovedAttribute("splittag", "urllib", "urllib.parse"),
    MovedAttribute("splituser", "urllib", "urllib.parse"),
    MovedAttribute("uses_fragment", "urlparse", "urllib.parse"),
    MovedAttribute("uses_netloc", "urlparse", "urllib.parse"),
    MovedAttribute("uses_params", "urlparse", "urllib.parse"),
    MovedAttribute("uses_query", "urlparse", "urllib.parse"),
    MovedAttribute("uses_relative", "urlparse", "urllib.parse"),
    ]
for attr in _urllib_parse_moved_attributes:
    setattr(Module_six_moves_urllib_parse, attr.name, attr)
del attr

Module_six_moves_urllib_parse._moved_attributes = _urllib_parse_moved_attributes

_importer._add_module(Module_six_moves_urllib_parse(__name__ + ".moves.urllib_parse"),
                      "moves.urllib_parse", "moves.urllib.parse")


class Module_six_moves_urllib_error(_LazyModule):
    """Lazy loading of moved objects in six.moves.urllib_error"""


_urllib_error_moved_attributes = [
    MovedAttribute("URLError", "urllib2", "urllib.error"),
    MovedAttribute("HTTPError", "urllib2", "urllib.error"),
    MovedAttribute("ContentTooShortError", "urllib", "urllib.error"),
    ]
for attr in _urllib_error_moved_attributes:
    setattr(Module_six_moves_urllib_error, attr.name, attr)
del attr

Module_six_moves_urllib_error._moved_attributes = _urllib_error_moved_attributes

_importer._add_module(Module_six_moves_urllib_error(__name__ + ".moves.urllib.error"),
                      "moves.urllib_error", "moves.urllib.error")


class Module_six_moves_urllib_request(_LazyModule):
    """Lazy loading of moved objects in six.moves.urllib_request"""


_urllib_request_moved_attributes = [
    MovedAttribute("urlopen", "urllib2", "urllib.request"),
    MovedAttribute("install_opener", "urllib2", "urllib.request"),
    MovedAttribute("build_opener", "urllib2", "urllib.request"),
    MovedAttribute("pathname2url", "urllib", "urllib.request"),
    MovedAttribute("url2pathname", "urllib", "urllib.request"),
    MovedAttribute("getproxies", "urllib", "urllib.request"),
    MovedAttribute("Request", "urllib2", "urllib.request"),
    MovedAttribute("OpenerDirector", "urllib2", "urllib.request"),
    MovedAttribute("HTTPDefaultErrorHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPRedirectHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPCookieProcessor", "urllib2", "urllib.request"),
    MovedAttribute("ProxyHandler", "urllib2", "urllib.request"),
    MovedAttribute("BaseHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPPasswordMgr", "urllib2", "urllib.request"),
    MovedAttribute("HTTPPasswordMgrWithDefaultRealm", "urllib2", "urllib.request"),
    MovedAttribute("AbstractBasicAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPBasicAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("ProxyBasicAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("AbstractDigestAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPDigestAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("ProxyDigestAuthHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPSHandler", "urllib2", "urllib.request"),
    MovedAttribute("FileHandler", "urllib2", "urllib.request"),
    MovedAttribute("FTPHandler", "urllib2", "urllib.request"),
    MovedAttribute("CacheFTPHandler", "urllib2", "urllib.request"),
    MovedAttribute("UnknownHandler", "urllib2", "urllib.request"),
    MovedAttribute("HTTPErrorProcessor", "urllib2", "urllib.request"),
    MovedAttribute("urlretrieve", "urllib", "urllib.request"),
    MovedAttribute("urlcleanup", "urllib", "urllib.request"),
    MovedAttribute("URLopener", "urllib", "urllib.request"),
    MovedAttribute("FancyURLopener", "urllib", "urllib.request"),
    MovedAttribute("proxy_bypass", "urllib", "urllib.request"),
    ]
for attr in _urllib_request_moved_attributes:
    setattr(Module_six_moves_urllib_request, attr.name, attr)
del attr

Module_six_moves_urllib_request._moved_attributes = _urllib_request_moved_attributes

_importer._add_module(Module_six_moves_urllib_request(__name__ + ".moves.urllib.request"),
                      "moves.urllib_request", "moves.urllib.request")


class Module_six_moves_urllib_response(_LazyModule):
    """Lazy loading of moved objects in six.moves.urllib_response"""


_urllib_response_moved_attributes = [
    MovedAttribute("addbase", "urllib", "urllib.response"),
    MovedAttribute("addclosehook", "urllib", "urllib.response"),
    MovedAttribute("addinfo", "urllib", "urllib.response"),
    MovedAttribute("addinfourl", "urllib", "urllib.response"),
    ]
for attr in _urllib_response_moved_attributes:
    setattr(Module_six_moves_urllib_response, attr.name, attr)
del attr

Module_six_moves_urllib_response._moved_attributes = _urllib_response_moved_attributes

_importer._add_module(Module_six_moves_urllib_response(__name__ + ".moves.urllib.response"),
                      "moves.urllib_response", "moves.urllib.response")


class Module_six_moves_urllib_robotparser(_LazyModule):
    """Lazy loading of moved objects in six.moves.urllib_robotparser"""


_urllib_robotparser_moved_attributes = [
    MovedAttribute("RobotFileParser", "robotparser", "urllib.robotparser"),
    ]
for attr in _urllib_robotparser_moved_attributes:
    setattr(Module_six_moves_urllib_robotparser, attr.name, attr)
del attr

Module_six_moves_urllib_robotparser._moved_attributes = _urllib_robotparser_moved_attributes

_importer._add_module(Module_six_moves_urllib_robotparser(__name__ + ".moves.urllib.robotparser"),
                      "moves.urllib_robotparser", "moves.urllib.robotparser")


class Module_six_moves_urllib(types.ModuleType):
    """Create a six.moves.urllib namespace that resembles the Python 3 namespace"""
    __path__ = []  # mark as package
    parse = _importer._get_module("moves.urllib_parse")
    error = _importer._get_module("moves.urllib_error")
    request = _importer._get_module("moves.urllib_request")
    response = _importer._get_module("moves.urllib_response")
    robotparser = _importer._get_module("moves.urllib_robotparser")

    def __dir__(self):
        return ['parse', 'error', 'request', 'response', 'robotparser']

_importer._add_module(Module_six_moves_urllib(__name__ + ".moves.urllib"),
                      "moves.urllib")


def add_move(move):
    """Add an item to six.moves."""
    setattr(_MovedItems, move.name, move)


def remove_move(name):
    """Remove item from six.moves."""
    try:
        delattr(_MovedItems, name)
    except AttributeError:
        try:
            del moves.__dict__[name]
        except KeyError:
            raise AttributeError("no such move, %r" % (name,))


if PY3:
    _meth_func = "__func__"
    _meth_self = "__self__"

    _func_closure = "__closure__"
    _func_code = "__code__"
    _func_defaults = "__defaults__"
    _func_globals = "__globals__"
else:
    _meth_func = "im_func"
    _meth_self = "im_self"

    _func_closure = "func_closure"
    _func_code = "func_code"
    _func_defaults = "func_defaults"
    _func_globals = "func_globals"


try:
    advance_iterator = next
except NameError:
    def advance_iterator(it):
        return it.next()
next = advance_iterator


try:
    callable = callable
except NameError:
    def callable(obj):
        return any("__call__" in klass.__dict__ for klass in type(obj).__mro__)


if PY3:
    def get_unbound_function(unbound):
        return unbound

    create_bound_method = types.MethodType

    Iterator = object
else:
    def get_unbound_function(unbound):
        return unbound.im_func

    def create_bound_method(func, obj):
        return types.MethodType(func, obj, obj.__class__)

    class Iterator(object):

        def next(self):
            return type(self).__next__(self)

    callable = callable
_add_doc(get_unbound_function,
         """Get the function out of a possibly unbound function""")


get_method_function = operator.attrgetter(_meth_func)
get_method_self = operator.attrgetter(_meth_self)
get_function_closure = operator.attrgetter(_func_closure)
get_function_code = operator.attrgetter(_func_code)
get_function_defaults = operator.attrgetter(_func_defaults)
get_function_globals = operator.attrgetter(_func_globals)


if PY3:
    def iterkeys(d, **kw):
        return iter(d.keys(**kw))

    def itervalues(d, **kw):
        return iter(d.values(**kw))

    def iteritems(d, **kw):
        return iter(d.items(**kw))

    def iterlists(d, **kw):
        return iter(d.lists(**kw))
else:
    def iterkeys(d, **kw):
        return iter(d.iterkeys(**kw))

    def itervalues(d, **kw):
        return iter(d.itervalues(**kw))

    def iteritems(d, **kw):
        return iter(d.iteritems(**kw))

    def iterlists(d, **kw):
        return iter(d.iterlists(**kw))

_add_doc(iterkeys, "Return an iterator over the keys of a dictionary.")
_add_doc(itervalues, "Return an iterator over the values of a dictionary.")
_add_doc(iteritems,
         "Return an iterator over the (key, value) pairs of a dictionary.")
_add_doc(iterlists,
         "Return an iterator over the (key, [values]) pairs of a dictionary.")


if PY3:
    def b(s):
        return s.encode("latin-1")
    def u(s):
        return s
    unichr = chr
    if sys.version_info[1] <= 1:
        def int2byte(i):
            return bytes((i,))
    else:
        # This is about 2x faster than the implementation above on 3.2+
        int2byte = operator.methodcaller("to_bytes", 1, "big")
    byte2int = operator.itemgetter(0)
    indexbytes = operator.getitem
    iterbytes = iter
    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO
else:
    def b(s):
        return s
    # Workaround for standalone backslash
    def u(s):
        return unicode(s.replace(r'\\', r'\\\\'), "unicode_escape")
    unichr = unichr
    int2byte = chr
    def byte2int(bs):
        return ord(bs[0])
    def indexbytes(buf, i):
        return ord(buf[i])
    def iterbytes(buf):
        return (ord(byte) for byte in buf)
    import StringIO
    StringIO = BytesIO = StringIO.StringIO
_add_doc(b, """Byte literal""")
_add_doc(u, """Text literal""")


if PY3:
    exec_ = getattr(moves.builtins, "exec")


    def reraise(tp, value, tb=None):
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

else:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")


    exec_("""def reraise(tp, value, tb=None):
    raise tp, value, tb
""")


print_ = getattr(moves.builtins, "print", None)
if print_ is None:
    def print_(*args, **kwargs):
        """The new-style print function for Python 2.4 and 2.5."""
        fp = kwargs.pop("file", sys.stdout)
        if fp is None:
            return
        def write(data):
            if not isinstance(data, basestring):
                data = str(data)
            # If the file has an encoding, encode unicode with it.
            if (isinstance(fp, file) and
                    isinstance(data, unicode) and
                        fp.encoding is not None):
                errors = getattr(fp, "errors", None)
                if errors is None:
                    errors = "strict"
                data = data.encode(fp.encoding, errors)
            fp.write(data)
        want_unicode = False
        sep = kwargs.pop("sep", None)
        if sep is not None:
            if isinstance(sep, unicode):
                want_unicode = True
            elif not isinstance(sep, str):
                raise TypeError("sep must be None or a string")
        end = kwargs.pop("end", None)
        if end is not None:
            if isinstance(end, unicode):
                want_unicode = True
            elif not isinstance(end, str):
                raise TypeError("end must be None or a string")
        if kwargs:
            raise TypeError("invalid keyword arguments to print()")
        if not want_unicode:
            for arg in args:
                if isinstance(arg, unicode):
                    want_unicode = True
                    break
        if want_unicode:
            newline = unicode("\n")
            space = unicode(" ")
        else:
            newline = "\n"
            space = " "
        if sep is None:
            sep = space
        if end is None:
            end = newline
        for i, arg in enumerate(args):
            if i:
                write(sep)
            write(arg)
        write(end)

_add_doc(reraise, """Reraise an exception.""")

if sys.version_info[0:2] < (3, 4):
    def wraps(wrapped, assigned=functools.WRAPPER_ASSIGNMENTS,
              updated=functools.WRAPPER_UPDATES):
        def wrapper(f):
            f = functools.wraps(wrapped)(f)
            f.__wrapped__ = wrapped
            return f
        return wrapper
else:
    wraps = functools.wraps

def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})


def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper

# Complete the moves implementation.
# This code is at the end of this module to speed up module loading.
# Turn this module into a package.
__path__ = []  # required for PEP 302 and PEP 451
__package__ = __name__  # see PEP 366 @ReservedAssignment
if globals().get("__spec__") is not None:
    __spec__.submodule_search_locations = []  # PEP 451 @UndefinedVariable
# Remove other six meta path importers, since they cause problems. This can
# happen if six is removed from sys.modules and then reloaded. (Setuptools does
# this for some reason.)
if sys.meta_path:
    for i, importer in enumerate(sys.meta_path):
        # Here's some real nastiness: Another "instance" of the six module might
        # be floating around. Therefore, we can't use isinstance() to check for
        # the six meta path importer, since the other six instance will have
        # inserted an importer with different class.
        if (type(importer).__name__ == "_SixMetaPathImporter" and
                    importer.name == __name__):
            del sys.meta_path[i]
            break
    del i, importer
# Finally, add the importer to the meta path import hook.
sys.meta_path.append(_importer)



# -*- coding: utf-8 -*-

import pytest
import warthog.api
import warthog.exceptions


@pytest.fixture
def exports():
    return set([item for item in dir(warthog.api) if not item.startswith('_')])


def test_public_exports(exports):
    declared = set(warthog.api.__all__)
    assert exports == declared, 'Exports and __all__ members should match'


def test_all_exceptions_imported(exports):
    errors = set([item for item in dir(warthog.exceptions) if item.endswith('Error') or item.endswith('Warning')])
    intersection = errors.intersection(exports)

    assert intersection == errors, "All available errors should be in warthog.api"




# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.cli
~~~~~~~~~~~

CLI interface for interacting with a load balancer using the Warthog client.
"""
import functools
import os
import os.path
import click
import warthog
import warthog.api
from .packages import six
import requests


def error_wrapper(func):
    """Decorator that coverts possible errors raised by the WarthogClient
    into instances of ClickExceptions so that they may be rendered automatically
    """

    # pylint: disable=missing-docstring
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except warthog.api.WarthogNoSuchNodeError as e:
            raise click.BadParameter("{0} doesn't appear to be a known node".format(e.server))
        except warthog.api.WarthogAuthFailureError as e:
            raise click.ClickException(
                "Authentication with the load balancer failed. The error was: {0}".format(e))
        except requests.ConnectionError as e:
            raise click.ClickException(
                "Connecting to the load balancer failed. The error was {0}".format(e))

    return wrapper


class WarthogClientFacade(object):
    """Wrapper around a :class:`warthog.client.WarthogClient` that coverts
    exceptions encountered into exceptions that click will handle automatically.
    """

    def __init__(self, client):
        self._client = client

    # pylint: disable=missing-docstring
    @error_wrapper
    def get_status(self, *args, **kwargs):
        return self._client.get_status(*args, **kwargs)

    # pylint: disable=missing-docstring
    @error_wrapper
    def get_connections(self, *args, **kwargs):
        return self._client.get_connections(*args, **kwargs)

    # pylint: disable=missing-docstring
    @error_wrapper
    def disable_server(self, *args, **kwargs):
        return self._client.disable_server(*args, **kwargs)

    # pylint: disable=missing-docstring
    @error_wrapper
    def enable_server(self, *args, **kwargs):
        return self._client.enable_server(*args, **kwargs)


@click.group()
@click.version_option(version=warthog.__version__)
@click.option(
    '--config',
    help='Path to a configuration file to use for the load balancer API.',
    type=click.Path(dir_okay=False))
@click.option(
    '--enable-platform-warning',
    help=('Enable warnings from underlying libraries when running on older Python '
          'versions known to cause intermittent failures of SSL/TLS connections.'),
    is_flag=True)
# pylint: disable=unused-argument
def main(config, enable_platform_warning):
    """Interact with a load balancer using the Warthog client."""
    # We don't actually do anything with the config file argument at this point.
    # The idea here is that we shouldn't be parsing the config file until we really
    # need it (like when we're creating a client instance). This allows us to display
    # help for subcommands without requiring the user to set up a config file first
    # (which would be really annoying).

    # Unless the user has specifically asked for this warning, we disable it because
    # it makes the CLI unusable on Python 2.6 or Python 2.7 < 2.7.9 (which is what CentOS
    # runs ATM).
    if not enable_platform_warning:
        disable_platform_warning()


def get_client(config):
    """Construct a new wrapped client based on the specified config file."""
    # Passing the config file unconditionally here since if the user hasn't
    # specified one it'll be None and the config loader will use the default
    # locations.
    loader = warthog.api.WarthogConfigLoader(config_file=config)

    try:
        # Expected errors that might be raised during parsing. These will
        # already have nice user-facing messages so we just reraise them as
        # BadParameter exceptions with the same message.
        loader.initialize()
    except warthog.api.WarthogConfigError as e:
        raise click.ClickException(six.text_type(e))

    settings = loader.get_settings()

    # Wrap the client in a facade that translates expected errors into
    # exceptions that click will render as error messages for the user.
    return WarthogClientFacade(warthog.api.WarthogClient(
        settings.scheme_host,
        settings.username,
        settings.password,
        verify=settings.verify))


def disable_platform_warning():
    """Disable the InsecurePlatformWarning emitted by urllib3. This is the
    default behavior unless the caller specifically asks for these warnings.

    See https://github.com/smarter-travel-media/warthog/issues/5
    """
    import warnings
    from requests.packages.urllib3.exceptions import InsecurePlatformWarning

    warnings.filterwarnings("ignore", category=InsecurePlatformWarning)


@click.command()
@click.argument('server')
@click.pass_context
def enable(ctx, server):
    """Enable a server by hostname."""
    client = get_client(ctx.parent.params['config'])
    if not client.enable_server(server):
        click.echo('{0} could not be enabled'.format(server))
        ctx.exit(1)


@click.command()
@click.argument('server')
@click.pass_context
def disable(ctx, server):
    """Disable a server by hostname."""
    client = get_client(ctx.parent.params['config'])
    if not client.disable_server(server):
        click.echo('{0} could not be disabled'.format(server))
        ctx.exit(1)


@click.command()
@click.argument('server')
@click.pass_context
def status(ctx, server):
    """Get the status of a server by hostname."""
    client = get_client(ctx.parent.params['config'])
    click.echo(client.get_status(server))


@click.command()
@click.argument('server')
@click.pass_context
def connections(ctx, server):
    """Get active connections to a server by hostname."""
    client = get_client(ctx.parent.params['config'])
    click.echo(client.get_connections(server))


@click.command('default-config')
def default_config():
    """Print a default configuration file."""
    click.echo(os.linesep.join([
        '[warthog]',
        'scheme_host = https://lb.example.com',
        'username = username',
        'password = password',
        'verify = yes',
        'ssl_version = TLSv1'
    ]))


@click.command('config-path')
def config_path():
    """Print the config file search PATH."""
    click.echo(os.linesep.join(warthog.api.DEFAULT_CONFIG_LOCATIONS))


main.add_command(enable)
main.add_command(disable)
main.add_command(status)
main.add_command(connections)
main.add_command(default_config)
main.add_command(config_path)



# -*- coding: utf-8 -*-

from click.testing import CliRunner

import pytest
import click
import requests

import warthog.cli
import warthog.exceptions


def test_main_no_command():
    runner = CliRunner()
    result = runner.invoke(warthog.cli.main)

    assert 0 == result.exit_code, "Expected zero exit code"


def test_main_bad_command():
    runner = CliRunner()
    result = runner.invoke(warthog.cli.main, args=['foo'])

    assert 0 != result.exit_code, "Expected non-zero exit code"


def test_main_default_config():
    runner = CliRunner()
    result = runner.invoke(warthog.cli.main, args=['default-config'])

    assert 0 == result.exit_code, 'Expected zero exit code'
    assert '[warthog]' in result.output, 'Expected "[warthog]" header in default config'


def test_main_config_path():
    runner = CliRunner()
    result = runner.invoke(warthog.cli.main, args=['config-path'])

    assert 0 == result.exit_code, 'Expected zero exit code'
    assert len(result.output), 'Expected at least some output'


def test_error_wrapper_no_such_node():
    @warthog.cli.error_wrapper
    def my_test_func(*_):
        raise warthog.exceptions.WarthogNoSuchNodeError(
            'No such server!', server='app1.example.com')

    with pytest.raises(click.BadParameter):
        my_test_func('something')


def test_error_wrapper_auth_failure():
    @warthog.cli.error_wrapper
    def my_test_func(*_):
        raise warthog.exceptions.WarthogAuthFailureError(
            'Authentication failed for some reason!')

    with pytest.raises(click.ClickException):
        my_test_func('something')


def test_error_wrapper_connection_error():
    @warthog.cli.error_wrapper
    def my_test_func(*_):
        raise requests.ConnectionError('No such LB host!')

    with pytest.raises(click.ClickException):
        my_test_func('something')




# -*- coding: utf-8 -*-


import ssl

import codecs
import pytest
import warthog.config
import warthog.exceptions
from warthog.packages.six.moves import configparser
import mock


class TestWarthogConfigLoader(object):
    def test_get_settings_not_parsed_yet(self):
        parser = mock.Mock(spec=warthog.config.WarthogConfigParser)
        loader = warthog.config.WarthogConfigLoader(config_file='warthog.ini', config_parser=parser)

        with pytest.raises(RuntimeError):
            loader.get_settings()

    def test_get_settings_success(self):
        parser = mock.Mock(spec=warthog.config.WarthogConfigParser)
        parser.parse.return_value = warthog.config.WarthogConfigSettings(
            scheme_host='https://lb.example.com',
            username='username',
            password='password',
            verify=None,
            ssl_version=None
        )

        loader = warthog.config.WarthogConfigLoader(config_file='warthog.ini', config_parser=parser)
        loader.initialize()
        settings = loader.get_settings()

        assert 'https://lb.example.com' == settings.scheme_host
        assert 'username' == settings.username
        assert 'password' == settings.password
        assert None is settings.verify
        assert None is settings.ssl_version


class TestWarthogConfigParser(object):
    def test_parse_no_config_file(self):
        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        open_impl = mock.MagicMock(spec=codecs.open)

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        with pytest.raises(warthog.exceptions.WarthogNoConfigFileError):
            parser.parse(None, 'utf-8', [])

    def test_parse_config_file_does_not_exist(self):
        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        open_impl = mock.MagicMock(spec=codecs.open)
        open_impl.side_effect = IOError("No such file!")

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        with pytest.raises(warthog.exceptions.WarthogNoConfigFileError):
            parser.parse('something.ini', 'utf-8', [])

    def test_parse_unicode_error(self):
        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        parser_impl.readfp.side_effect = UnicodeDecodeError('ascii', b'\x80abc', 0, 1, 'ordinal not in range(128)')
        open_impl = mock.MagicMock(spec=codecs.open)

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        with pytest.raises(warthog.exceptions.WarthogMalformedConfigFileError):
            parser.parse('something.ini', 'ascii', [])

    def test_parse_no_section(self):
        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        parser_impl.get.side_effect = configparser.NoSectionError('warthog')
        open_impl = mock.MagicMock(spec=codecs.open)

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        with pytest.raises(warthog.exceptions.WarthogMalformedConfigFileError):
            parser.parse('something.ini', 'utf-8', [])

    def test_parse_no_option(self):
        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        parser_impl.get.side_effect = configparser.NoOptionError('warthog', 'scheme_host')
        open_impl = mock.MagicMock(spec=codecs.open)

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        with pytest.raises(warthog.exceptions.WarthogMalformedConfigFileError):
            parser.parse('something.ini', 'utf-8', [])

    def test_parse_success(self):
        def has_option_impl(section, option):
            return True

        def get_impl(section, option):
            if option == 'scheme_host':
                return 'https://lb.example.com'
            if option == 'username':
                return 'user'
            if option == 'password':
                return 'pass'
            if option == 'ssl_version':
                return 'TLSv1'
            raise ValueError('No such option ' + option)

        def getboolean_impl(section, option):
            return True

        parser_impl = mock.Mock(spec=configparser.SafeConfigParser)
        parser_impl.has_option.side_effect = has_option_impl
        parser_impl.get.side_effect = get_impl
        parser_impl.getboolean.side_effect = getboolean_impl

        open_impl = mock.MagicMock(spec=codecs.open)

        parser = warthog.config.WarthogConfigParser(parser_impl=parser_impl, open_impl=open_impl)

        settings = parser.parse('something.ini', 'utf-8', [])

        assert 'https://lb.example.com' == settings.scheme_host
        assert 'user' == settings.username
        assert 'pass' == settings.password
        assert True == settings.verify
        assert ssl.PROTOCOL_TLSv1 == settings.ssl_version


class TestWarthogConfigResolver(object):
    def test_call_explicit_config_file(self):
        obj = warthog.config.WarthogConfigFileResolver([])

        assert '/tmp/something.ini' == obj('/tmp/something.ini')[0]

    def test_call_no_explicit_file_default_exists(self):
        default_file = '/etc/warthog.ini'
        exists = lambda x: x == default_file
        obj = warthog.config.WarthogConfigFileResolver(['/tmp/something.ini', default_file], exists_impl=exists)

        assert default_file == obj(None)[0]

    def test_call_no_explicit_file_default_does_not_exist(self):
        exists = lambda x: False
        obj = warthog.config.WarthogConfigFileResolver(['/tmp/something.ini', '/tmp/else.ini'], exists_impl=exists)

        assert None == obj(None)[0]


class FakeModule(object):
    """Class to wrap ``object`` so that we can set attributes on it
    but still get attribute errors for properties that aren't set, unlike
    :class:`mock.Mock` which will just hand back a mock instance
    """


def test_parse_ssl_version_none_input():
    assert None is warthog.config.parse_ssl_version(None)


def test_parse_ssl_version_blank_string():
    assert None is warthog.config.parse_ssl_version(u"    ")


def test_parse_ssl_version_empty_string():
    assert None is warthog.config.parse_ssl_version(u'')


def test_parse_ssl_version_valid_version():
    mod = FakeModule()
    mod.PROTOCOL_SSLv23 = 2
    mod.PROTOCOL_TLSv1 = 3
    mod.PROTOCOL_TLSv1_1 = 4
    mod.PROTOCOL_TLSv1_2 = 5

    assert mod.PROTOCOL_SSLv23 == warthog.config.parse_ssl_version(u'SSLv23', mod)
    assert mod.PROTOCOL_TLSv1 == warthog.config.parse_ssl_version(u'TLSv1', mod)
    assert mod.PROTOCOL_TLSv1_1 == warthog.config.parse_ssl_version(u'TLSv1_1', mod)
    assert mod.PROTOCOL_TLSv1_2 == warthog.config.parse_ssl_version(u'TLSv1_2', mod)


def test_parse_ssl_version_unsupported_version():
    mod = FakeModule()
    mod.PROTOCOL_TLSv1 = 3

    with pytest.raises(ValueError):
        warthog.config.parse_ssl_version(u'SSLv2', mod)

    with pytest.raises(ValueError):
        warthog.config.parse_ssl_version(u'TLSv1_1', mod)

    with pytest.raises(ValueError):
        warthog.config.parse_ssl_version(u'TLSv1_2', mod)


def test_parse_ssl_version_malformed_version():
    mod = FakeModule()
    mod.PROTOCOL_TLSv1 = 3

    with pytest.raises(ValueError):
        warthog.config.parse_ssl_version(u'SOMETHING', mod)




# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.core
~~~~~~~~~~~~

Basic building blocks for authentication and interaction with a load balancer.
"""

import logging

import warthog.exceptions
# pylint: disable=import-error,no-name-in-module
from .packages.six.moves import urllib


STATUS_ENABLED = 'enabled'

STATUS_DISABLED = 'disabled'

STATUS_DOWN = 'down'

ERROR_CODE_GRACEFUL_SHUTDOWN = 67174416

ERROR_CODE_INVALID_SESSION = 1009

ERROR_CODE_NO_SUCH_SERVER = 67174402

TRANSIENT_ERRORS = frozenset([
    ERROR_CODE_GRACEFUL_SHUTDOWN
])

_API_VERSION = 'v2'

_ACTION_AUTHENTICATE = 'authenticate'

_ACTION_ENABLE = _ACTION_DISABLE = 'slb.server.update'

_ACTION_STATUS = _ACTION_STATISTICS = 'slb.server.fetchStatistics'

_ACTION_CLOSE_SESSION = 'session.close'


def get_log():
    """Get the :class:`logging.Logger` instance used by the Warthog library.

    :return: Logger for the entire library
    :rtype: logging.Logger
    """
    return logging.getLogger('warthog')


class _ErrorHandlerMixin(object):
    """Mixin class for translating error responses to WarthogApiError instances."""

    def _extract_error(self, default_cls, message, response):
        """Return a WarthogApiError instance with the given error message appropriate
        to the given API response, falling back to an instance of the default class.

        If the default class is a node-specific error, the name of the current node
        will be provided with the ``server`` keyword argument.

        The ``default_cls`` must be a subclass of ``WarthogApiError``.
        """
        server = getattr(self, '_server', None)
        msg, code = _extract_error_message(response)

        if code == ERROR_CODE_INVALID_SESSION:
            return warthog.exceptions.WarthogInvalidSessionError(
                message, api_msg=msg, api_code=code)
        if code == ERROR_CODE_NO_SUCH_SERVER:
            return warthog.exceptions.WarthogNoSuchNodeError(
                message, api_msg=msg, api_code=code, server=server)
        # If this is a node specific error type include the value of the '_server'
        # attribute of the class that this is being invoked from as a keyword arg.
        if issubclass(default_cls, warthog.exceptions.WarthogNodeError):
            return default_cls(message, api_msg=msg, api_code=code, server=server)
        return default_cls(message, api_msg=msg, api_code=code)


class SessionStartCommand(_ErrorHandlerMixin):
    """Command to authenticate with the load balancer and start a new session
    to be used by subsequent commands.

    This class is thread safe.
    """
    _logger = get_log()

    def __init__(self, transport, scheme_host, username, password):
        """Set the transport layer and necessary credentials to authenticate with
        the load balancer.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring username: Name of the user to authenticate with.
        :param basestring password: Password for the user to authenticate with.
        """
        self._transport = transport
        self._scheme_host = scheme_host
        self._username = username
        self._password = password

    def send(self):
        """Make an authentication request and return the session token that should
        be included in all subsequent requests to the load balancer.

        :return: The session token that should be used for all subsequent requests
            made to the load balancer.
        :rtype: unicode
        :raises warthog.exceptions.WarthogAuthFailureError: If the authentication
            failed for some reason. The exception will contain an error message and
            error code that provides more detail about the failure. Common reasons
            for this error include using invalid username or password.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_AUTHENTICATE)
        params['username'] = self._username
        params['password'] = self._password

        self._logger.debug('Making session start POST request to %s', url)
        response = self._transport.post(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if 'session_id' not in json:
            raise self._extract_error(
                warthog.exceptions.WarthogAuthFailureError,
                'Authentication failure with {0}'.format(self._scheme_host),
                json['response'])

        return json['session_id']


class _AuthenticatedCommand(object):
    """Base class for making requests to the load balancer using an existing session
    ID from a previous :class:`SessionStartCommand` request.

    :ivar requests.Session _transport:
    :ivar basestring _scheme_host:
    :ivar basestring _session_id:
    """
    _logger = get_log()

    def __init__(self, transport, scheme_host, session_id):
        """Set the requests transport layer, scheme and host of the load balancer,
        and existing session ID to use for authentication.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring session_id: Session ID from a previous authentication request
            made to the load balancer.
        """
        self._transport = transport
        self._scheme_host = scheme_host
        self._session_id = session_id

    def send(self):
        """Abstract method for making a request to the load balancer API and parsing
        the result and returning any meaningful information (implementation specific).
        """
        raise NotImplementedError()


class SessionEndCommand(_AuthenticatedCommand, _ErrorHandlerMixin):
    """Command for ending a previously authenticated session with the load balancer.

    This class is thread safe.
    """

    def send(self):
        """Close an existing session and return ``True`` if closing it was successful.

        :return: True if the current session could be closed
        :rtype: bool
        :raises warthog.exceptions.WarthogInvalidSessionError: If the load balancer
            did not recognize the session this command is being run as part of.
        :raises warthog.exceptions.WarthogAuthCloseError: If the session could not be
            closed. This is usually the result of the session ID being invalid or the
            session already being closed before this command is run.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_CLOSE_SESSION, self._session_id)

        self._logger.debug('Making session close POST request to %s', url)
        response = self._transport.post(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if json['response']['status'] == 'fail':
            raise self._extract_error(
                warthog.exceptions.WarthogAuthCloseError,
                'Could not close session {0} on {1}'.format(
                    self._session_id, self._scheme_host),
                json['response'])

        return json['response']['status'] == 'OK'


class NodeEnableCommand(_AuthenticatedCommand, _ErrorHandlerMixin):
    """Command to mark a particular server as enabled.

    This class is thread safe.
    """

    def __init__(self, transport, scheme_host, session_id, server):
        """Set the requests transport layer, scheme and host of the load balancer,
        existing session ID to use for authentication, and hostname of the server
        to enable.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring session_id: Session ID from a previous authentication request
            made to the load balancer.
        :param basestring server: Host name of the server to enable.
        """
        super(NodeEnableCommand, self).__init__(transport, scheme_host, session_id)
        self._server = server

    def send(self):
        """Mark a server as 'enabled' at the node level and return ``True`` if it was
        successfully enabled.

        :return: True if the server was marked as enabled
        :rtype: bool
        :raises warthog.exceptions.WarthogInvalidSessionError: If the load balancer
            did not recognize the session this command is being run as part of.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the server was not
            recognized by the load balancer.
        :raises warthog.exceptions.WarthogNodeEnableError: If the server could
            not be enabled for any other reason.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_ENABLE, self._session_id)
        params['name'] = self._server
        params['server'] = self._server
        params['status'] = 1

        self._logger.debug('Making node enable POST request for %s', self._server)
        response = self._transport.post(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if json['response']['status'] == 'fail':
            raise self._extract_error(
                warthog.exceptions.WarthogNodeEnableError,
                'Could not enable node {0}'.format(self._server),
                json['response'])

        return json['response']['status'] == 'OK'


class NodeDisableCommand(_AuthenticatedCommand, _ErrorHandlerMixin):
    """Command to mark a particular server as disabled.

    This class is thread safe.
    """

    def __init__(self, transport, scheme_host, session_id, server):
        """Set the requests transport layer, scheme and host of the load balancer,
        existing session ID to use for authentication, and hostname of the server
        to disable.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring session_id: Session ID from a previous authentication request
            made to the load balancer.
        :param basestring server: Host name of the server to disable.
        """
        super(NodeDisableCommand, self).__init__(transport, scheme_host, session_id)
        self._server = server

    def send(self):
        """Mark a server as 'disabled' at the node level and return ``True`` if it
        was successfully disabled.

        :return: True if the server was marked as disabled
        :rtype: bool
        :raises warthog.exceptions.WarthogInvalidSessionError: If the load balancer
            did not recognize the session this command is being run as part of.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the server was not
            recognized by the load balancer.
        :raises warthog.exceptions.WarthogNodeDisableError: If the server could
            not be disabled for any other reason.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_ENABLE, self._session_id)
        params['name'] = self._server
        params['server'] = self._server
        params['status'] = 0

        self._logger.debug('Making node disable POST request for %s', self._server)
        response = self._transport.post(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if json['response']['status'] == 'fail':
            raise self._extract_error(
                warthog.exceptions.WarthogNodeDisableError,
                'Could not disable node {0}'.format(self._server),
                json['response'])

        return json['response']['status'] == 'OK'


class NodeStatusCommand(_AuthenticatedCommand, _ErrorHandlerMixin):
    """Command to get the current status ('enabled', 'disabled', 'down') of a particular
    server.

    This class is thread safe.
    """

    def __init__(self, transport, scheme_host, session_id, server):
        """Set the requests transport layer, scheme and host of the load balancer,
        existing session ID to use for authentication, and hostname of the server
        to get the status of.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring session_id: Session ID from a previous authentication request
            made to the load balancer.
        :param basestring server: Host name of the server to get the status of.
        """
        super(NodeStatusCommand, self).__init__(transport, scheme_host, session_id)
        self._server = server

    def send(self):
        """Get the current status of a server at the node level and return one of the
        ``STATUS_ENABLED``, ``STATUS_DISABLED``, ``STATUS_DOWN`` constants.

        :return: The status of the server as a constant string
        :rtype: basestring
        :raises warthog.exceptions.WarthogInvalidSessionError: If the load balancer
            did not recognize the session this command is being run as part of.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the server was not
            recognized by the load balancer.
        :raises warthog.exceptions.WarthogNodeStatusError: If the status of the server
            could not be determined for any other reason.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_STATUS, self._session_id)
        params['name'] = self._server

        self._logger.debug('Making node status GET request for %s', self._server)
        response = self._transport.get(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if 'server_stat' not in json:
            raise self._extract_error(
                warthog.exceptions.WarthogNodeStatusError,
                'Could not get status of {0}'.format(self._server),
                json['response'])

        status = json['server_stat']['status']
        if status == 0:
            return STATUS_DISABLED
        if status == 1:
            return STATUS_ENABLED
        if status == 2:
            return STATUS_DOWN

        raise warthog.exceptions.WarthogNodeStatusError(
            'Unknown status of {0}: status={1}'.format(self._server, status))


class NodeActiveConnectionsCommand(_AuthenticatedCommand, _ErrorHandlerMixin):
    """Command to get the number of active connections to a particular server.

    This class is thread safe.
    """

    def __init__(self, transport, scheme_host, session_id, server):
        """Set the requests transport layer, scheme and host of the load balancer,
        existing session ID to use for authentication, and hostname of the server
        to get active connections for.

        :param requests.Session transport: Configured requests session instance to
            use for making HTTP or HTTPS requests to the load balancer API.
        :param basestring scheme_host: Scheme and hostname of the load balancer to use for
            making API requests. E.g. 'https://lb.example.com' or 'http://10.1.2.3'.
        :param basestring session_id: Session ID from a previous authentication request
            made to the load balancer.
        :param basestring server: Host name of the server to get active connections for.
        """
        super(NodeActiveConnectionsCommand, self).__init__(transport, scheme_host, session_id)
        self._server = server

    def send(self):
        """Get the current number of active connections for a node as an int.

        :return: The number of active connections for a node across all ports
        :rtype: int
        :raises warthog.exceptions.WarthogInvalidSessionError: If the load balancer
            did not recognize the session this command is being run as part of.
        :raises warthog.exceptions.WarthogNoSuchNodeError: If the server was not
            recognized by the load balancer.
        :raises warthog.exceptions.WarthogNodeStatusError: If the number of active
            connections to the server could not be determined for any other reason.
        """
        url = _get_base_url(self._scheme_host)
        params = _get_base_query_params(_ACTION_STATISTICS, self._session_id)
        params['name'] = self._server

        self._logger.debug('Making active connection count GET request for %s', self._server)
        response = self._transport.get(url, params=params)
        self._logger.debug(response.text)
        json = response.json()

        if 'server_stat' not in json:
            raise self._extract_error(
                warthog.exceptions.WarthogNodeStatusError,
                'Could not get status of {0}'.format(self._server),
                json['response'])

        return json['server_stat']['cur_conns']


def _extract_error_message(response):
    """Get a two element tuple of the form ``msg, code`` where ``msg`` is an
    error message returned by the API and ``code`` is a numeric code associated
    with that particular error.
    """
    if response['status'] == 'fail':
        return response['err']['msg'].strip(), response['err']['code']
    raise ValueError(
        "Unexpected response format from request: {0}".format(response))


def _get_base_url(scheme_host):
    """Get a URL to API of the load balancer, not including any query string
    parameters.
    """
    return urllib.parse.urljoin(
        scheme_host, 'services/rest/{version}/'.format(version=_API_VERSION))


def _get_base_query_params(action, session_id=None):
    """Get a dictionary of query string parameters to pass to the load balancer
    API based on the given action and optional session ID.
    """
    params = {
        'format': 'json',
        'method': action
    }

    if session_id is not None:
        params['session_id'] = session_id
    return params



# -*- coding: utf-8 -*-

import ssl

import warthog.transport


def test_get_transport_factory_no_verify():
    factory = warthog.transport.get_transport_factory(verify=False)
    session = factory()
    adapter = session.get_adapter('https://lb.example.com')

    assert isinstance(
        adapter, warthog.transport.VersionedSSLAdapter), 'Did not get expected HTTPS adapter'
    assert not session.verify, 'Expected verify to be disabled on session'


def test_get_transport_factory_alternate_ssl_version():
    factory = warthog.transport.get_transport_factory(ssl_version=ssl.PROTOCOL_SSLv3)
    session = factory()
    adapter = session.get_adapter('https://lb.example.com')

    assert ssl.PROTOCOL_SSLv3 == adapter.ssl_version, 'Did not get expected SSL version'


def test_get_transport_factory_with_defaults():
    factory = warthog.transport.get_transport_factory(verify=None, ssl_version=None)
    session = factory()
    adapter = session.get_adapter('https://lb.example.com')

    assert warthog.transport.DEFAULT_SSL_VERSION == adapter.ssl_version, 'Did not get default TLS version'
    assert warthog.transport.DEFAULT_CERT_VERIFY == session.verify, 'Did not get default verify setting'



# -*- coding: utf-8 -*-

import logging

import pytest
import mock
import warthog.client
import warthog.core
import warthog.exceptions


SCHEME_HOST = 'https://lb.example.com'


@pytest.fixture
def commands(
        start_cmd,
        end_cmd,
        status_cmd,
        conn_cmd,
        enable_cmd,
        disable_cmd):
    factory = mock.Mock(spec=warthog.client.CommandFactory)
    factory.get_session_start.return_value = start_cmd
    factory.get_session_end.return_value = end_cmd
    factory.get_server_status.return_value = status_cmd
    factory.get_enable_server.return_value = enable_cmd
    factory.get_disable_server.return_value = disable_cmd
    factory.get_active_connections.return_value = conn_cmd
    return factory


@pytest.fixture
def start_cmd():
    return mock.Mock(spec=warthog.core.SessionStartCommand)


@pytest.fixture
def end_cmd():
    return mock.Mock(spec=warthog.core.SessionEndCommand)


@pytest.fixture
def status_cmd():
    return mock.Mock(spec=warthog.core.NodeStatusCommand)


@pytest.fixture
def conn_cmd():
    return mock.Mock(spec=warthog.core.NodeActiveConnectionsCommand)


@pytest.fixture
def enable_cmd():
    return mock.Mock(spec=warthog.core.NodeEnableCommand)


@pytest.fixture
def disable_cmd():
    return mock.Mock(spec=warthog.core.NodeDisableCommand)


@pytest.fixture
def client():
    return mock.Mock(spec=warthog.client.WarthogClient)


def test_session_context_enter_yields_session(commands, start_cmd):
    start_cmd.send.return_value = '1234'

    context = warthog.client.session_context(SCHEME_HOST, 'user', 'password', commands)
    session = context.__enter__()

    assert '1234' == session, 'Did not get expected session ID'
    commands.get_session_start.assert_called_once_with(SCHEME_HOST, 'user', 'password')
    assert start_cmd.send.called, 'Expected session start .send() to be called'


def test_session_context_exit_closes_previous_session(commands, start_cmd, end_cmd):
    start_cmd.send.return_value = '1234'

    context = warthog.client.session_context(SCHEME_HOST, 'user', 'password', commands)
    context.__enter__()
    context.__exit__(None, None, None)

    commands.get_session_end.assert_called_once_with(SCHEME_HOST, '1234')
    assert end_cmd.send.called, 'Expected session end .send() to be called'


def test_session_context_exception_in_context_propagated(commands, start_cmd):
    start_cmd.send.return_value = '1234'
    context = warthog.client.session_context(SCHEME_HOST, 'user', 'password', commands)

    with pytest.raises(RuntimeError):
        with context:
            raise RuntimeError("AHH!")


class TestWarthogClient(object):
    def test_get_status(self, commands, start_cmd, end_cmd, status_cmd):
        start_cmd.send.return_value = '1234'
        status_cmd.send.return_value = 'down'

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        status = client.get_status('app1.example.com')

        assert 'down' == status, 'Did not get expected status'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_get_connections(self, commands, start_cmd, end_cmd, conn_cmd):
        start_cmd.send.return_value = '1234'
        conn_cmd.send.return_value = 42

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        connections = client.get_connections('app1.example.com')

        assert 42 == connections, 'Did not get expected active connections'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_disable_server_no_active_connections(self, commands, start_cmd, end_cmd,
                                                  status_cmd, conn_cmd, disable_cmd):
        start_cmd.send.return_value = '1234'
        disable_cmd.send.return_value = True
        conn_cmd.send.return_value = 0
        status_cmd.send.return_value = 'disabled'

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        disabled = client.disable_server('app1.example.com')

        assert disabled, 'Server did not end up disabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_disable_server_with_active_connections(self, commands, start_cmd, end_cmd,
                                                    status_cmd, conn_cmd, disable_cmd):
        start_cmd.send.return_value = '1234'
        disable_cmd.send.return_value = True
        conn_cmd.send.return_value = [42, 3, 0]
        status_cmd.send.return_value = 'disabled'

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        disabled = client.disable_server('app1.example.com')

        assert disabled, 'Server did not end up disabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_disable_server_never_gets_disabled(self, commands, start_cmd, end_cmd,
                                                status_cmd, conn_cmd, disable_cmd):
        start_cmd.send.return_value = '1234'
        disable_cmd.send.return_value = True
        conn_cmd.send.return_value = 42
        status_cmd.send.return_value = 'enabled'

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        disabled = client.disable_server('app1.example.com')

        assert not disabled, 'Server ended up disabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_enable_server_in_graceful_shutdown(self, commands, start_cmd, end_cmd,
                                                status_cmd, enable_cmd):
        graceful_error = warthog.exceptions.WarthogNodeEnableError(
            '', api_code=warthog.core.ERROR_CODE_GRACEFUL_SHUTDOWN, api_msg='')

        start_cmd.send.return_value = '1234'
        enable_cmd.send.side_effect = [graceful_error, graceful_error, True]
        status_cmd.send.return_value = 'enabled'

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        enabled = client.enable_server('app1.example.com')

        assert enabled, 'Server did not end up enabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_enable_server_in_graceful_shutdown_switching_to_down(self, commands, start_cmd, end_cmd,
                                                                  status_cmd, enable_cmd):
        graceful_error = warthog.exceptions.WarthogNodeEnableError(
            '', api_code=warthog.core.ERROR_CODE_GRACEFUL_SHUTDOWN, api_msg='')

        start_cmd.send.return_value = '1234'
        enable_cmd.send.side_effect = [graceful_error, graceful_error, True]
        status_cmd.send.side_effect = ['down', 'down', 'enabled', 'enabled']

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        enabled = client.enable_server('app1.example.com')

        assert enabled, 'Server did not end up enabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_enable_server_no_graceful_shutdown_switching_to_down(self, commands, start_cmd, end_cmd,
                                                                  status_cmd, enable_cmd):
        start_cmd.send.return_value = '1234'
        enable_cmd.send.return_value = True
        status_cmd.send.side_effect = ['down', 'down', 'enabled', 'enabled']

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        enabled = client.enable_server('app1.example.com')

        assert enabled, 'Server did not end up enabled'
        assert end_cmd.send.called, 'Session end .send() did not get called'

    def test_enable_server_never_comes_out_of_graceful_shutdown(self, commands, start_cmd, end_cmd,
                                                                enable_cmd):
        graceful_error = warthog.exceptions.WarthogNodeEnableError(
            '', api_code=warthog.core.ERROR_CODE_GRACEFUL_SHUTDOWN, api_msg='')

        start_cmd.send.return_value = '1234'
        enable_cmd.send.side_effect = [graceful_error, graceful_error]

        client = warthog.client.WarthogClient(
            SCHEME_HOST, 'user', 'password', wait_interval=0.1, commands=commands)

        with pytest.raises(warthog.exceptions.WarthogNodeEnableError):
            client.enable_server('app1.example.com', max_retries=1)

        assert end_cmd.send.called, 'Session end .send() did not get called'



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#


import codecs

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import warthog


AUTHOR = 'Smarter Travel'
EMAIL = ''
DESCRIPTION = 'Simple client for A10 load balancers'
URL = 'https://github.com/smarter-travel-media/warthog'
LICENSE = 'MIT'
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: MIT License",
    "Operating System :: POSIX",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Topic :: System :: Installation/Setup",
    "Topic :: System :: Software Distribution"
]

REQUIREMENTS = [
    'click',
    'requests'
]

with codecs.open('README.rst', 'r', 'utf-8') as handle:
    LONG_DESCRIPTION = handle.read()

setup(
    name='warthog',
    version=warthog.__version__,
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author_email=EMAIL,
    classifiers=CLASSIFIERS,
    license=LICENSE,
    url=URL,
    install_requires=REQUIREMENTS,
    zip_safe=True,
    packages=['warthog', 'warthog.packages'],
    entry_points="""
        [console_scripts]
        warthog=warthog.cli:main
    """)



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.config
~~~~~~~~~~~~~~

Load and parse configuration for a client from an INI-style file.
"""

import collections
import threading
import ssl
import sys

import codecs
import os.path
import warthog.exceptions
from .packages import six
# pylint: disable=import-error
from .packages.six.moves import configparser

# List of locations (from most preferred to least preferred) that will
# be searched for a configuration file. These locations are typically
# only searched when an explicit configuration file is not used.
DEFAULT_CONFIG_LOCATIONS = [
    os.path.join('/etc', 'warthog', 'warthog.ini'),
    os.path.join('/etc', 'warthog.ini'),
    os.path.join(sys.prefix, 'etc', 'warthog', 'warthog.ini'),
    os.path.join(sys.prefix, 'etc', 'warthog.ini'),
    os.path.join(os.path.expanduser('~'), '.warthog.ini'),
    os.path.join(os.getcwd(), 'warthog.ini')
]


# By default, we assume that the configuration file is in UTF-8 unless
# the caller indicates it is in some other encoding.
DEFAULT_CONFIG_ENCODING = 'utf-8'


# Simple immutable struct to hold configuration information for a WarthogClient
WarthogConfigSettings = collections.namedtuple(
    'WarthogConfigSettings', ['scheme_host', 'username', 'password', 'verify', 'ssl_version'])


class WarthogConfigLoader(object):
    """Load and parse configuration from an INI-style WarthogClient configuration file.

    If a specific configuration file is given during construction, this file will
    be used instead of checking the multiple possible default locations for configuration
    files. The default locations to be checked are an ordered list of paths contained
    in :data:`DEFAULT_CONFIG_LOCATIONS`.

    .. note::

        When checking for configuration files in default locations, each file will only
        be checked to see if it exists. It will not be checked to see if the file is
        readable or correctly formatted.

    This class is thread safe.

    .. versionadded:: 0.4.0

    .. versionchanged:: 0.6.0
        Loading, parsing, and access of configuration settings is now thread safe.

    .. versionchanged:: 0.6.0
        The .parse_configuration() method has been removed and the functionality has
        been split into the .initialize() and .get_settings() methods.

    .. versionchanged:: 0.10.0
        The :meth:`WarthogConfigLoader.__init__` method no longer directly takes a standard
        library INI parser as an option parameter, instead it now takes a WarthogConfigParser
        instance as an optional parameter.

    .. versionchanged:: 0.10.0
        See :doc:`changes` or :doc:`cli` for details about the changes to configuration
        file format.
    """

    def __init__(self, config_file=None, encoding=None, path_resolver=None, config_parser=None):
        """Optionally, set a specific configuration file, the encoding of the file, resolver
        to determine the configuration file to use, and custom configuration parser implementation.

        By default, multiple locations will be checked for a configuration file, the file
        is assumed to use UTF-8 encoding, and an

        :param str|unicode config_file: Optional explicit path to a configuration file
            to use.
        :param str|unicode encoding: Encoding to use for reading the configuration file.
            Default is UTF-8
        :param callable path_resolver: Callable that accepts a single argument (the explicit
            configuration file path to use) and determines what configuration file to use. It
            is typically only necessary to set this parameter for unit testing purposes.
        :param WarthogConfigParser config_parser: Optional configuration parser to use for
            reading and parsing the expected INI format for a Warthog configuration file. It
            is typically only necessary to set this parameter for unit testing purposes.
        """
        self._config_file = config_file
        self._encoding = encoding if encoding is not None else DEFAULT_CONFIG_ENCODING

        self._path_resolver = path_resolver if path_resolver is not None else \
            WarthogConfigFileResolver(DEFAULT_CONFIG_LOCATIONS)

        self._parser = config_parser if config_parser is not None else \
            WarthogConfigParser(configparser.SafeConfigParser())

        self._lock = threading.RLock()
        self._settings = None

    def initialize(self):
        """Load and parse a configuration an INI-style configuration file.

        The values parsed will be stored as a :class:`WarthogConfigSettings` instance that
        may be accessed with the :meth:`get_settings` method.

        .. versionadded:: 0.6.0

        .. versionchanged:: 0.8.0
            Errors locating or parsing configuration files now result in Warthog-specific
            exceptions (:class:`warthog.exceptions.WarthogConfigError`) instead of
            `ValueError`, `IOError`, or `RuntimeError`.

        :return: Fluent interface
        :rtype: WarthogConfigLoader
        :raises warthog.exceptions.WarthogNoConfigFileError: If no explicit configuration file
            was given and there were no configuration files in any of the default locations
            checked or if the configuration file could not be opened or read for some
            reason.
        :raises warthog.exceptions.WarthogMalformedConfigFileError: If the configuration
            file was malformed such has missing the required 'warthog' section or any of
            the expected values. See the :doc:`cli` section for more information about the
            expected configuration settings.
        """
        with self._lock:
            config_file, checked = self._path_resolver(self._config_file)
            self._settings = self._parser.parse(config_file, self._encoding, checked)
        return self

    def get_settings(self):
        """Get previously loaded and parsed configuration settings, raise an exception
        if the settings have not already been loaded and parsed.

        .. versionadded:: 0.6.0

        :return: Struct of configuration settings for the Warthog client
        :rtype: WarthogConfigSettings
        :raises RuntimeError: If a configuration file has not already been loaded and
            parsed.
        """
        with self._lock:
            if self._settings is None:
                raise RuntimeError(
                    "Configuration file must be loaded and parsed before "
                    "settings can be used (via the .initialize() method)")
            return self._settings


def parse_ssl_version(version_str, ssl_module=None):
    """Get the :mod:`ssl` protocol constant that represents the given version
    string if it exists, raising an error if the version string is malformed or
    does not correspond to a supported protocol.

    :param unicode version_str: Version string to resolve to a protocol
    :param module ssl_module: SSL module to get the protocol constant from
    :return: The ssl module protocol constant or ``None``
    :raises ValueError: If the version string did not match any known versions
        of SSL or TLS
    """
    if version_str is None:
        return None

    version_str = version_str.strip()
    if not version_str:
        return None

    ssl_module = ssl_module if ssl_module is not None else ssl

    # Get a list of all the 'PROTOCOL' constants in the SSL module, and
    # strip the 'PROTOCOL_' prefix. This is the set of supported SSL or
    # TLS versions that we'll compare the user input against.
    supported = set([const.replace('PROTOCOL_', '', 1) for const in dir(ssl_module)
                     if const.startswith('PROTOCOL_')])

    if version_str in supported:
        return getattr(ssl_module, 'PROTOCOL_' + version_str)

    raise ValueError(
        "Unsupported SSL/TLS version '" + version_str + "'. Supported: " + ', '.join(supported))


class WarthogConfigFileResolver(object):
    """Callable that returns a tuple of the form $path, $searched where
     $path is the configuration file that should be used or None (if there
     were no suitable files found) and $searched is a list of all paths that
     were checked to find a suitable file.

     If a configuration file was explicitly provided, it will be used without
     checking any of the default locations. If no file was explicitly provided
     each of the default locations will be checked to find one configuration
     file that exists.

     If no suitable file could be found, the callable will return the tuple
     ``None``, $searched where $searched is the list of locations checked.
    """

    def __init__(self, default_locations, exists_impl=None):
        self._default_locations = default_locations
        self._exists_impl = exists_impl if exists_impl is not None else os.path.exists

    def __call__(self, path):
        if path is not None:
            return path, [path]

        for default in self._default_locations:
            if self._exists_impl(default):
                return default, self._default_locations

        return None, self._default_locations


class WarthogConfigParser(object):
    """Facade for a standard library INI file parser that parses the expected
    configuration values for configuring a :class:`warthog.client.WarthogClient`
    instance.

    All configuration values are expected to be in the ``warthog`` section of
    the INI file. The ``ssl_version`` and ``verify`` values are not required, all
    others are.

    This class is not thread safe.
    """

    def __init__(self, parser_impl, open_impl=None):
        """Set the underlying standard library INI parser to use for reading
        Warthog configuration settings.

        :param configparser.RawConfigParser parser_impl: INI file parser to use
            for parsing Warthog configuration settings.
        :param callable open_impl: Open method for opening the configuration
            file. The method is expected to have the same signature as
            :func:`codes.open`. Callers should only need to supply this for unit
            testing purposes.
        """
        self._parser_impl = parser_impl
        self._open_impl = open_impl if open_impl is not None else codecs.open

    def _load_file(self, path, encoding, checked):
        """Open and load the configuration file at the given path."""
        if path is None:
            raise warthog.exceptions.WarthogNoConfigFileError(
                "No configuration file was specified. Please set a "
                "configuration file or ensure that a configuration "
                "file exists in one of the default locations checked",
                locations_checked=checked)

        try:
            with self._open_impl(path, 'r', encoding=encoding) as handle:
                self._parser_impl.readfp(handle)
        except IOError as e:
            six.reraise(
                warthog.exceptions.WarthogNoConfigFileError,
                warthog.exceptions.WarthogNoConfigFileError(
                    "The configuration file does not exist or could not read. "
                    "Please make sure {0} exists and can be read by the current "
                    "user. Original error message: {1}".format(path, e),
                    locations_checked=checked),
                sys.exc_info()[2])
        except UnicodeError as e:
            six.reraise(
                warthog.exceptions.WarthogMalformedConfigFileError,
                warthog.exceptions.WarthogMalformedConfigFileError(
                    "The configuration file {0} doesn't seem to be correctly encoded "
                    "{1} text. Please ensure that the file is valid text. Original "
                    "error message: {2}".format(path, encoding, e)),
                sys.exc_info()[2])

    def _get_ssl_version(self, section, option):
        """Get the specified TLS version in the config file or None."""
        if self._parser_impl.has_option(section, option):
            return parse_ssl_version(self._parser_impl.get(section, option))
        return None

    def _get_verify(self, section, option):
        """Get the certificate verify option in the config file or None."""
        if self._parser_impl.has_option(section, option):
            return self._parser_impl.getboolean(section, option)
        return None

    def _parse_file(self):
        """Parse the opened configuration file and return the results as a namedtuple."""
        try:
            scheme_host = self._parser_impl.get('warthog', 'scheme_host')
            username = self._parser_impl.get('warthog', 'username')
            password = self._parser_impl.get('warthog', 'password')
            verify = self._get_verify('warthog', 'verify')
            ssl_version = self._get_ssl_version('warthog', 'ssl_version')
        except configparser.NoSectionError as e:
            raise warthog.exceptions.WarthogMalformedConfigFileError(
                "The configuration file seems to be missing a '{0}' section. Please "
                "make sure this section exists".format(e.section), missing_section=e.section)
        except configparser.NoOptionError as e:
            raise warthog.exceptions.WarthogMalformedConfigFileError(
                "The configuration file seems to be missing the '{0}' option. Please "
                "make sure this option exists".format(e.option), missing_option=e.option)

        return WarthogConfigSettings(
            scheme_host=scheme_host,
            username=username,
            password=password,
            verify=verify,
            ssl_version=ssl_version)

    def parse(self, path, encoding, checked):
        """Attempt to open and parse the configuration file at the given
        path.

        :param str|unicode path: Path to the configuration file to parse.
        :param str|unicode encoding: Encoding to use when opening the file.
        :param list checked: List of the various locations checked before
            deciding on the configuration file to open.
        :return: The parsed configuration settings to use for creating a client.
        :rtype: WarthogConfigSettings
        """

        self._load_file(path, encoding, checked)
        return self._parse_file()



# -*- coding: utf-8 -*-
#
# Warthog - Simple client for A10 load balancers
#
# Copyright 2014-2015 Smarter Travel
#
# Available under the MIT license. See LICENSE for details.
#

"""
warthog.api
~~~~~~~~~~~

Publicly importable API for the Warthog client and library.
"""

from .core import (
    STATUS_DISABLED,
    STATUS_DOWN,
    STATUS_ENABLED)

from .client import (
    CommandFactory,
    WarthogClient)

from .config import (
    WarthogConfigLoader,
    WarthogConfigSettings,
    DEFAULT_CONFIG_ENCODING,
    DEFAULT_CONFIG_LOCATIONS)

from .transport import get_transport_factory

from .exceptions import (
    WarthogError,
    WarthogApiError,
    WarthogAuthCloseError,
    WarthogAuthFailureError,
    WarthogInvalidSessionError,
    WarthogNodeError,
    WarthogNodeDisableError,
    WarthogNodeEnableError,
    WarthogNodeStatusError,
    WarthogNoSuchNodeError,
    WarthogConfigError,
    WarthogMalformedConfigFileError,
    WarthogNoConfigFileError)


__all__ = [
    # warthog.core
    'STATUS_DISABLED',
    'STATUS_DOWN',
    'STATUS_ENABLED',

    # warthog.client
    'CommandFactory',
    'WarthogClient',

    # warthog.config
    'WarthogConfigLoader',
    'WarthogConfigSettings',
    'DEFAULT_CONFIG_ENCODING',
    'DEFAULT_CONFIG_LOCATIONS',

    # warthog.transport
    'get_transport_factory',

    # warthog.exceptions
    'WarthogError',
    'WarthogApiError',
    'WarthogAuthCloseError',
    'WarthogAuthFailureError',
    'WarthogInvalidSessionError',
    'WarthogNodeError',
    'WarthogNodeDisableError',
    'WarthogNodeEnableError',
    'WarthogNodeStatusError',
    'WarthogNoSuchNodeError',
    'WarthogConfigError',
    'WarthogMalformedConfigFileError',
    'WarthogNoConfigFileError'
]



# -*- coding: utf-8 -*-

import warthog.exceptions


class TestWarthogApiError(object):
    def test_str(self):
        err = warthog.exceptions.WarthogApiError('Oh no!', api_msg='Something bad', api_code=1234)
        str_err = str(err)

        assert 'Oh no!' in str_err, 'Did not see expected message in error'
        assert 'Something bad' in str_err, 'Did not see expected API message in error'
        assert '1234' in str_err, 'Did not see expected API code in error'
